==> hdsync-1.0.tar.gz: hdsync-1.0/src/hdsync_cli.cpp <==
/* HDSync
   
 (c) 2011-2012 Montevideo / Time Based Arts (aka NIMk.nl)
     2011-2012 Denis Roio <jaromil@dyne.org>

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Affero General Public License as
     published by the Free Software Foundation, either version 3 of the
     License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU Affero General Public License for more details.
     
     You should have received a copy of the GNU Affero General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>

*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
//#include <netdb.h>

#include <errno.h>

#include <sys/types.h>
#include <ifaddrs.h>
#include <netinet/in.h> 
#include <arpa/inet.h>

#include <avremote.h>
#include <zmq.h>
#include <zmq_utils.h>

// our exit codes are shell style: 1 is error, 0 is success
#define ERR 1

// uncomment to debug
#define DEBUG 1

// network settings
#define ETHDEV "eth0"
#define PGMPORT "5555"


// commandline switches
char filename[512];
char server[512];
int port = 0;
int dryrun = 0;
int chanID = 0;
int chanTOT = 0;


// info on peers
upnp_t **peers;


// own IPv4 address
char IPv4[24];

// 0mq
void *zcontext;
void *sock_in = NULL;
void *sock_shout = NULL;
char pgmaddr[64];
zmq_msg_t request;

// we use only getopt, no _long
static const char *short_options = "-hvts:p:";

void cmdline(int argc, char **argv);


int get_address();

// return 1 on matched string
// 0 otherwise
int get_state(const char *match);


int shout(const char *message, int flags) {
  int res;
  //  zmq_connect(sock_out, pgmaddr);
  zmq_msg_init_size (&request, strlen(message));
  memcpy (zmq_msg_data (&request), message, strlen(message));
  printf ("Shouting \"%s\"\n", message);
  res = zmq_send (sock_shout, &request, flags);
  zmq_msg_close (&request);
  //  zmq_close(sock_out);
  return(res);
}

// extract a string from a message received:
// PREFIX;MESSAGE;
// message should be a char of 512 bytes 
int expect(const char *prefix, char *message, int flags) {
  zmq_msg_t reply;
  char res[512];
  char *p, *pp;
  int pfxlen = strlen(prefix);

  //  printf("Expecting prefix %s\n",prefix);

  // listen for answers
  zmq_msg_init (&reply);
  zmq_recv (sock_in, &reply, flags);
  // printf ("Received response\n");
  snprintf(res,511,"%s",(char*)zmq_msg_data(&reply));
  zmq_msg_close (&reply);
  
  // parse the response and if ACK insert it in uniq list
  if( strncmp(res,prefix,pfxlen) == 0) {
    // quick and secure string parsing
    p = res; while(*p != ';') p++; p++;
    pp = p; while(*pp != ';') pp++; *pp='\0';
    
    // *p has the ip string
    snprintf(message,511,"%s",p);
    //    printf("Expected prefix %s found %s\n",prefix,message);
    return(0);
  }
  
  return(1);
}

int handshake() {
  char message[256];
  int listindex = 1;
  int listmax = 0;
  char res[512];
  int c;

  listmax = chanTOT; // number of desired peers
  printf("looking to connect %u peers\n",listmax);

  // allocate space for listener array
  for(c=1; c<listmax; c++)
    peers[c] = create_upnp();

  if(chanID==1) { // first channel is the one to offer

    ////////
    // OFFER
    //////////////////////////////
    printf ("Offering on %s from address %s\n",pgmaddr, IPv4);

    // offer subscribes to all broadcasts by default
    // somehow listening to ACK only doesn't works
    snprintf(message,255,"OFFER;%s;",IPv4);
    while(1) {
      
      shout(message, 0);

      if( expect("ACK",res, 0) == 0) {
	//	printf("Got ACK from %s\n",res);
	// parse the response and if ACK insert it in uniq list
	for(c=0; c<listindex; c++)
	  if(strncmp(peers[c]->hostname, res, 16) == 0)
	    break; // found a duplicate
	if(c==listindex) { // there was no duplicate
	  // parse and insert hostname and port
	  char *p = res; while(*p != ':') p++; *p='\0';
	  snprintf(peers[c]->hostname, 16, "%s", res);
	  sscanf(p+1, "%u", &peers[c]->port);
	  printf("New listener: %s : %u\n",peers[c]->hostname, peers[c]->port);
	  listindex++;
	}
	
	if(listindex==listmax) { // goal reached
	  printf("Sending ready signals to listeners\n");
	  ///////////////////////////////////////////////
	  shout("READY", 0);
	  break; // break offer loop
	} // if listmax reached
	
      } // if ACK response

      zmq_sleep(1);

    }	// offer while loop
    
  } else { // all other channels listen

      /////////
      // LISTEN

      printf ("Listening on %s from address %s\n",pgmaddr, peers[0]->hostname);


      snprintf(message,255,"ACK;%s:%u;",peers[0]->hostname,peers[0]->port);


      while (1) {
        //  Wait for offer
	while( expect("OFFER", res, 0) !=0); // blocking
	// reply with ACK
	shout(message, 0);
	
	while( expect("READY", res, 0) !=0); // blocking
	printf ("Received ready signal\n");
	break; // quit loop on ready signal
	
      }

  }
    
  return(1);
}

int main(int argc, char **argv) {
  char tmp[512];
  int c;

  //////////
  ///// INIT

  cmdline(argc, argv);

  if(!filename[0]) {
    fprintf(stderr,"not enough args specified on commandline, see help.\n");
    exit(ERR);
  }

  fprintf(stderr,"HDSync starting for channel %u of %u\n",chanID, chanTOT);
  fprintf(stderr,"will sync and play video: %s\n", filename);

  if( ! get_address() )  {
    fprintf(stderr,"Error getting own address.\n");
    exit(ERR);
  }

  // we put ourselves in first
  peers = (upnp_t**) calloc(chanTOT, sizeof(upnp_t*));
  peers[0] = create_upnp();
  snprintf(peers[0]->hostname, 16, "%s", IPv4);
  peers[0]->port = port;

  if(!dryrun) {
    if ( connect_upnp (peers[0]) < 0 ) {
      fprintf(stderr,"can't connect to local UPNP: operation aborted.\n");
      // free all peers array
      for(c=0; c<chanTOT; c++)
	free_upnp(peers[c]);
      exit(ERR);
    } else {
      fprintf(stderr,"UPNP connected to %s:%u\n",
	      peers[0]->hostname, peers[0]->port);
      close(peers[0]->sockfd);
      peers[0]->sockfd = 0;
    }
  }


  // initialize 0mq
  zcontext = zmq_init (1);

  // create broadcast socket
  sock_shout = zmq_socket (zcontext, ZMQ_PUB);
  zmq_connect(sock_shout, pgmaddr);

  // create input socket
  sock_in = zmq_socket (zcontext, ZMQ_SUB);
  zmq_bind (sock_in, pgmaddr);
  zmq_setsockopt(sock_in,ZMQ_SUBSCRIBE,"",0);


  /// END OF INIT
  ////////////////  


  ////////////////////
  if(! handshake()) {
    // no error return so far, just running endless untill contact is established
    fprintf(stderr,"error establishing contact with all desired channels, aborting.\n");
    exit(ERR);
  }

  printf("Handshake successful\n");

  //////////////////////
  if(dryrun) { // DRYRUN
    printf("Dry run\n");
    
    // testing sync with no real playback
    zmq_sleep(1);

    if(chanID==1) { // offer should send sync

      printf("Prepare to offer sync\n");
      
      zmq_sleep(3);

      shout("SYNC;;", 0);

    } else { // listener should prepare to receive sync
      printf("Prepare to receive sync\n");

      while( expect("SYNC", tmp, 0) !=0); // blocking

    }
    
    //////////////////////

  } else { // RUN FOR REAL, FUN FOR REAL

    //////////////////////
  
    // load, play and pause in sequence
    // break and reopen connection in between
    

    // stop player
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Stop","");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    zmq_sleep(1);

    // load video into player
    connect_upnp(peers[0]);
    render_uri_meta(peers[0],filename);
    render_upnp(peers[0],"SetAVTransportURI", peers[0]->meta);
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;
    
    zmq_sleep(1);

    // play
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Play","<Speed>1</Speed>");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    // offer checks if all are playing
    // if(chanID==1) {
    //   int playing = 0;
    //   while(playing < listmax)

    zmq_sleep(5);
    
    connect_upnp(peers[0]);
    snprintf(tmp,511,"<Unit>REL_TIME</Unit><Target>%s</Target>",filename);
    render_upnp(peers[0],"Seek","<Unit>REL_TIME</Unit><Target>00:00:00</Target>");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    // pause
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Pause","");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;
        
    // prepare upnp play message
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Play","<Speed>1</Speed>");
    
    if(chanID==1) { // offer should send sync

      zmq_sleep(2);

      shout("SYNC", 0); // send sync signal

    } else {
      
      //  wait for sync
      while( expect("SYNC", tmp, 0) !=0); // blocking

    }

    // start player
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);


    zmq_sleep(1);

    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    
  } // if(test|real)

  while( ! get_state("STOPPED") ) zmq_sleep(1);

  fprintf(stderr,"Software shutdown\n");

  zmq_close (sock_shout);

  zmq_close (sock_in);

  // somehow offer blocks here
  printf("free 0mq context\n");
  zmq_term (zcontext);

  printf("free peer array\n");
  // free all peers array
  for(c=0; c<chanTOT; c++) {
    if(peers[c] != NULL)
      free_upnp(peers[c]);
  }
  free(peers);

  fprintf(stderr,"Done.\n");
  exit(0);
}


void cmdline(int argc, char **argv) {
  filename[0] = 0;
  server[0] = 0;

  int res;
  do {
    res = getopt(argc, argv, short_options);
    switch(res) {
    case 'h':
      fprintf(stderr,
	      "%s %s - prepare and start synced playback of video across devices\n"
	      "\n"
	      " Copyright (C) 2011-2012 Jaromil @ Dyne.org , License GNU AGPL v3+\n"
	      " This is free software: you are free to change and redistribute it.\n"
	      " The latest HDSync sourcecode is published on <%s>\n"
	      "\n"
	      "Syntax: hdsync [options] channel_id channel_total filename\n"
	      "\n"
	      "Options:\n"
	      "\n"
	      " -s          network address or hostname of the media server\n"
	      " -p          port on which the UPNP AVTransport daemon is listening\n"
	      " -t          dry run to test without a server (print out rendered xml)\n"
	      "\n"
	      " -h          print this help\n"
	      " -v          version information for this tool\n"
	      "\n"
	      "For more informations on HDSync see <http://syncstarter.org>\n"
	      "Please report bugs on <http://bugs.dyne.org>.\n",
	      PACKAGE, VERSION, PACKAGE_URL);
      exit(0);

    case 'v':
      fprintf(stderr,"%s - simple commandline tool to send AVTransport commands over UPNP\n"
	      "version %s distributed by Montevideo/NIMk\n"
	      "Copyright (C) 2011-2012 Jaromil @ Dyne.org, License GNU AGPL v3+\n"
	      "This is free software: you are free to change and redistribute it\n",
	      PACKAGE, VERSION);
      exit(0);
      
    case 's':
      snprintf(server,511,"%s",optarg);
      break;

    case 'p':
      sscanf (optarg, "%u", &port);
      break;

    case 't':
      dryrun = 1;
      break;

    case '?':
      fprintf(stderr,"unrecognized option: %s\n",optarg);
      break;
      
    case 1:
      if(!chanID) {
	sscanf(optarg,"%u",&chanID);
      } else if(!chanTOT) {
	sscanf(optarg,"%u",&chanTOT);
      } else if(!filename[0]) {
	snprintf(filename,511,"%s",optarg);
      }
      break;
    default:
      break;
    }

  } while(res != -1);
}

int get_address() {
    // discover own IP address
    struct ifaddrs * ifAddrStruct=NULL;
    struct ifaddrs * ifa=NULL;
    void * tmpAddrPtr=NULL;

    printf("Handshake on LAN\n");

    getifaddrs(&ifAddrStruct);

    pgmaddr[0] = NULL; // zero if iface not found

    for (ifa = ifAddrStruct; ifa != NULL; ifa = ifa->ifa_next) {
        if (ifa ->ifa_addr->sa_family==AF_INET) { // check it is IP4
            // is a valid IP4 Address
            tmpAddrPtr=&((struct sockaddr_in *)ifa->ifa_addr)->sin_addr;
            char addressBuffer[INET_ADDRSTRLEN];
            inet_ntop(AF_INET, tmpAddrPtr, addressBuffer, INET_ADDRSTRLEN);
            // printf("%s IP Address %s\n", ifa->ifa_name, addressBuffer);
	    if( strcmp(ifa->ifa_name, ETHDEV) == 0) {
	      strncpy(IPv4, addressBuffer, 16);
	      snprintf(pgmaddr,63,"pgm://%s:%s",ifa->ifa_name, PGMPORT);
	    }
	}
    }
    if (ifAddrStruct!=NULL) freeifaddrs(ifAddrStruct);
    
    if(!pgmaddr[0]) {
      printf("Error: ethernet interface %s is not configured\n",ETHDEV);
      return(0);
    }
    return(1);
}



#define MAX 256 // max XML string size
// returns a pointer to the current position in haystack
// for subsequent calls
char *extract_xml(char *dest, char *haystack, const char *needle) {
  char *p, *pp;

  p = strstr(haystack, needle);
  if (!p) {
    fprintf(stderr, "parser cannot find \"%s\" in:\n%s\n",needle, haystack);
    return NULL;
  }
  // as you can see, i had a lot of fun with pointers when i was a kid
  pp = p; do pp+=1; while (*pp != '>'); pp++;
  p = pp; do p+=1; while (*p != '<'); *p = 0;

  snprintf(dest,MAX-1,"%s",pp);

  return(p+1);

}
int get_state(const char *match) {
  char state[MAX];
  char status[MAX];
  char speed[MAX];
  char *p;

  connect_upnp(peers[0]);
  render_upnp(peers[0],"GetTransportInfo","");
  send_upnp(peers[0]);
  recv_upnp(peers[0], 1000);
  close(peers[0]->sockfd);
  peers[0]->sockfd = 0;

  p = extract_xml(state, peers[0]->res, "CurrentTransportState"); if(!p) return(0);
  p = extract_xml(status, p, "CurrentTransportStatus");
  p = extract_xml(speed, p, "CurrentSpeed");

  //  fprintf(stderr,"#\tstate\tstatus\tspeed\n");
  //  fprintf(stderr,"TInfo:\t%s\t%s\t%s\n", state, status, speed);

  if( strcmp(match, state) == 0) return(1);
  
  return(0);
}

==> hdsync-1.1.tar.gz: hdsync-1.1/src/hdsync_cli.cpp <==
/* HDSync
   
 (c) 2011-2012 Montevideo / Time Based Arts (aka NIMk.nl)
     2011-2012 Denis Roio <jaromil@dyne.org>

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Affero General Public License as
     published by the Free Software Foundation, either version 3 of the
     License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU Affero General Public License for more details.
     
     You should have received a copy of the GNU Affero General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>

*/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
//#include <netdb.h>

#include <errno.h>

#include <sys/types.h>
#include <ifaddrs.h>
#include <netinet/in.h> 
#include <arpa/inet.h>

#include <avremote.h>
#include <zmq.h>
#include <zmq_utils.h>

// our exit codes are shell style: 1 is error, 0 is success
#define ERR 1

// uncomment to debug
#define DEBUG 1

// network settings
#define PGMPORT "5555"


// commandline switches
char filename[512];
char server[512];
char pgmstatic[256];
int port = 0;
int dryrun = 0;
int chanID = 0;
int chanTOT = 0;
char ethiface[64];


// info on peers
upnp_t **peers;


// own IPv4 address
char IPv4[24];

// 0mq
void *zcontext;
void *sock_in = NULL;
void *sock_shout = NULL;
char pgmaddr[256];
zmq_msg_t request;

// we use only getopt, no _long
static const char *short_options = "-hvta:i:s:p:";

void cmdline(int argc, char **argv);


int get_address();

// return 1 on matched string
// 0 otherwise
int get_state(const char *match);


int shout(const char *message, int flags) {
  int res;
  //  zmq_connect(sock_out, pgmaddr);
  zmq_msg_init_size (&request, strlen(message));
  memcpy (zmq_msg_data (&request), message, strlen(message));
  printf ("Shouting \"%s\"\n", message);
  res = zmq_send (sock_shout, &request, flags);
  zmq_msg_close (&request);
  //  zmq_close(sock_out);
  return(res);
}

// extract a string from a message received:
// PREFIX;MESSAGE;
// message should be a char of 512 bytes 
int expect(const char *prefix, char *message, int flags) {
  zmq_msg_t reply;
  char res[512];
  char *p, *pp;
  int pfxlen = strlen(prefix);

  //  printf("Expecting prefix %s\n",prefix);

  // listen for answers
  zmq_msg_init (&reply);
  zmq_recv (sock_in, &reply, flags);
  // printf ("Received response\n");
  snprintf(res,511,"%s",(char*)zmq_msg_data(&reply));
  zmq_msg_close (&reply);
  
  // parse the response and if ACK insert it in uniq list
  if( strncmp(res,prefix,pfxlen) == 0) {
    // quick and secure string parsing
    p = res; while(*p != ';') p++; p++;
    pp = p; while(*pp != ';') pp++; *pp='\0';
    
    // *p has the ip string
    snprintf(message,511,"%s",p);
    //    printf("Expected prefix %s found %s\n",prefix,message);
    return(0);
  }
  
  return(1);
}

int handshake() {
  char message[256];
  int listindex = 1;
  int listmax = 0;
  char res[512];
  int c;

  listmax = chanTOT; // number of desired peers
  printf("looking to connect %u peers\n",listmax);

  // allocate space for listener array
  for(c=1; c<listmax; c++)
    peers[c] = create_upnp();

  if(chanID==1) { // first channel is the one to offer

    ////////
    // OFFER
    //////////////////////////////
    printf ("Offering on %s from address %s\n",pgmaddr, IPv4);

    // offer subscribes to all broadcasts by default
    // somehow listening to ACK only doesn't works
    snprintf(message,255,"OFFER;%s;",IPv4);
    while(1) {
      
      shout(message, 0);

      if( expect("ACK",res, 0) == 0) {
	//	printf("Got ACK from %s\n",res);
	// parse the response and if ACK insert it in uniq list
	for(c=0; c<listindex; c++)
	  if(strncmp(peers[c]->hostname, res, 16) == 0)
	    break; // found a duplicate
	if(c==listindex) { // there was no duplicate
	  // parse and insert hostname and port
	  char *p = res; while(*p != ':') p++; *p='\0';
	  snprintf(peers[c]->hostname, 16, "%s", res);
	  sscanf(p+1, "%u", &peers[c]->port);
	  printf("New listener: %s : %u\n",peers[c]->hostname, peers[c]->port);
	  listindex++;
	}
	
	if(listindex==listmax) { // goal reached
	  printf("Sending ready signals to listeners\n");
	  ///////////////////////////////////////////////
	  shout("READY", 0);
	  break; // break offer loop
	} // if listmax reached
	
      } // if ACK response

      zmq_sleep(1);

    }	// offer while loop
    
  } else { // all other channels listen

      /////////
      // LISTEN

      printf ("Listening on %s from address %s\n",pgmaddr, peers[0]->hostname);


      snprintf(message,255,"ACK;%s:%u;",peers[0]->hostname,peers[0]->port);


      while (1) {
        //  Wait for offer
	while( expect("OFFER", res, 0) !=0); // blocking
	// reply with ACK
	shout(message, 0);
	
	while( expect("READY", res, 0) !=0); // blocking
	printf ("Received ready signal\n");
	break; // quit loop on ready signal
	
      }

  }
    
  return(1);
}

int main(int argc, char **argv) {
  char tmp[512];
  int c;

  //////////
  ///// INIT

  cmdline(argc, argv);

  if(!filename[0]) {
    fprintf(stderr,"not enough args specified on commandline, see help.\n");
    exit(ERR);
  }

  fprintf(stderr,"HDSync starting for channel %u of %u\n",chanID, chanTOT);
  fprintf(stderr,"will sync and play video: %s\n", filename);

  if( ! get_address() )  {
    fprintf(stderr,"Error getting own address.\n");
    exit(ERR);
  }

  // we put ourselves in first
  peers = (upnp_t**) calloc(chanTOT, sizeof(upnp_t*));
  peers[0] = create_upnp();
  snprintf(peers[0]->hostname, 16, "%s", IPv4);
  peers[0]->port = port;

  if(!dryrun) {
    if ( connect_upnp (peers[0]) < 0 ) {
      fprintf(stderr,"can't connect to local UPNP: operation aborted.\n");
      // free all peers array
      for(c=0; c<chanTOT; c++)
	free_upnp(peers[c]);
      exit(ERR);
    } else {
      fprintf(stderr,"UPNP connected to %s:%u\n",
	      peers[0]->hostname, peers[0]->port);
      close(peers[0]->sockfd);
      peers[0]->sockfd = 0;
    }
  }


  // initialize 0mq
  zcontext = zmq_init (1);

  // create broadcast socket
  sock_shout = zmq_socket (zcontext, ZMQ_PUB);
  zmq_connect(sock_shout, pgmaddr);

  // create input socket
  sock_in = zmq_socket (zcontext, ZMQ_SUB);
  zmq_bind (sock_in, pgmaddr);
  zmq_setsockopt(sock_in,ZMQ_SUBSCRIBE,"",0);


  /// END OF INIT
  ////////////////  


  ////////////////////
  if(! handshake()) {
    // no error return so far, just running endless untill contact is established
    fprintf(stderr,"error establishing contact with all desired channels, aborting.\n");
    exit(ERR);
  }

  printf("Handshake successful\n");

  //////////////////////
  if(dryrun) { // DRYRUN
    printf("Dry run\n");
    
    // testing sync with no real playback
    zmq_sleep(1);

    if(chanID==1) { // offer should send sync

      printf("Prepare to offer sync\n");
      
      zmq_sleep(3);

      shout("SYNC;;", 0);

    } else { // listener should prepare to receive sync
      printf("Prepare to receive sync\n");

      while( expect("SYNC", tmp, 0) !=0); // blocking

    }
    
    //////////////////////

  } else { // RUN FOR REAL, FUN FOR REAL

    //////////////////////
  
    // load, play and pause in sequence
    // break and reopen connection in between
    

    // stop player
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Stop","");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    zmq_sleep(1);

    // load video into player
    connect_upnp(peers[0]);
    render_uri_meta(peers[0],filename);
    render_upnp(peers[0],"SetAVTransportURI", peers[0]->meta);
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;
    
    zmq_sleep(1);

    // play
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Play","<Speed>1</Speed>");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    // offer checks if all are playing
    // if(chanID==1) {
    //   int playing = 0;
    //   while(playing < listmax)

    zmq_sleep(5);
    
    connect_upnp(peers[0]);
    snprintf(tmp,511,"<Unit>REL_TIME</Unit><Target>%s</Target>",filename);
    render_upnp(peers[0],"Seek","<Unit>REL_TIME</Unit><Target>00:00:00</Target>");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    // pause
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Pause","");
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);
    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;
        
    // prepare upnp play message
    connect_upnp(peers[0]);
    render_upnp(peers[0],"Play","<Speed>1</Speed>");
    
    if(chanID==1) { // offer should send sync

      zmq_sleep(2);

      shout("SYNC", 0); // send sync signal

    } else {
      
      //  wait for sync
      while( expect("SYNC", tmp, 0) !=0); // blocking

    }

    // start player
    send_upnp(peers[0]);
    recv_upnp(peers[0], 1000);


    zmq_sleep(1);

    close(peers[0]->sockfd);
    peers[0]->sockfd = 0;

    
  } // if(test|real)

  while( ! get_state("STOPPED") ) zmq_sleep(1);

  fprintf(stderr,"Software shutdown\n");

  zmq_close (sock_shout);

  zmq_close (sock_in);

  // somehow offer blocks here
  printf("free 0mq context\n");
  zmq_term (zcontext);

  printf("free peer array\n");
  // free all peers array
  for(c=0; c<chanTOT; c++) {
    if(peers[c] != NULL)
      free_upnp(peers[c]);
  }
  free(peers);

  fprintf(stderr,"Done.\n");
  exit(0);
}


void cmdline(int argc, char **argv) {
  filename[0] = 0;
  server[0] = 0;
  pgmstatic[0] = 0;

  int res;
  do {
    res = getopt(argc, argv, short_options);
    switch(res) {
    case 'h':
      fprintf(stderr,
	      "%s %s - prepare and start synced playback of video across devices\n"
	      "\n"
	      " Copyright (C) 2011-2012 Jaromil @ Dyne.org , License GNU AGPL v3+\n"
	      " This is free software: you are free to change and redistribute it.\n"
	      " The latest HDSync sourcecode is published on <%s>\n"
	      "\n"
	      "Syntax: hdsync [options] channel_id channel_total filename\n"
	      "\n"
	      "Options:\n"
	      "\n"
	      " -s          network address or hostname of the media server\n"
	      " -p          port on which the UPNP AVTransport daemon is listening\n"
	      " -t          dry run to test without a server (print out rendered xml)\n"
	      " -a          manually specified pgm:// address for broadcast range\n"
	      " -i          network interface to use (default: eth0)\n"
	      "\n"
	      " -h          print this help\n"
	      " -v          version information for this tool\n"
	      "\n"
	      "For more informations on HDSync see <http://syncstarter.org>\n"
	      "Please report bugs on <http://bugs.dyne.org>.\n",
	      PACKAGE, VERSION, PACKAGE_URL);
      exit(0);

    case 'v':
      fprintf(stderr,"%s - simple commandline tool to send AVTransport commands over UPNP\n"
	      "version %s distributed by Montevideo/NIMk\n"
	      "Copyright (C) 2011-2012 Jaromil @ Dyne.org, License GNU AGPL v3+\n"
	      "This is free software: you are free to change and redistribute it\n",
	      PACKAGE, VERSION);
      exit(0);
      
    case 's':
      snprintf(server,511,"%s",optarg);
      break;

    case 'a':
      snprintf(pgmstatic,255,"%s",optarg);
      break;

    case 'i':
      snprintf(ethiface,63,"%s",optarg);
      break;

    case 'p':
      sscanf (optarg, "%u", &port);
      break;

    case 't':
      dryrun = 1;
      break;


    case '?':
      fprintf(stderr,"unrecognized option: %s\n",optarg);
      break;
      
    case 1:
      if(!chanID) {
	sscanf(optarg,"%u",&chanID);
      } else if(!chanTOT) {
	sscanf(optarg,"%u",&chanTOT);
      } else if(!filename[0]) {
	snprintf(filename,511,"%s",optarg);
      }
      break;
    default:
      break;
    }

  } while(res != -1);
}

int get_address() {
    // discover own IP address
    struct ifaddrs * ifAddrStruct=NULL;
    struct ifaddrs * ifa=NULL;
    void * tmpAddrPtr=NULL;

    printf("Handshake on LAN\n");

    getifaddrs(&ifAddrStruct);

    pgmaddr[0] = NULL; // zero if iface not found

    for (ifa = ifAddrStruct; ifa != NULL; ifa = ifa->ifa_next) {
        if (ifa ->ifa_addr->sa_family==AF_INET) { // check it is IP4
            // is a valid IP4 Address
            tmpAddrPtr=&((struct sockaddr_in *)ifa->ifa_addr)->sin_addr;
            char addressBuffer[INET_ADDRSTRLEN];
            inet_ntop(AF_INET, tmpAddrPtr, addressBuffer, INET_ADDRSTRLEN);
            // printf("%s IP Address %s\n", ifa->ifa_name, addressBuffer);
	    if( strcmp(ifa->ifa_name, ethiface) == 0) {
	      strncpy(IPv4, addressBuffer, 16);
	      if( pgmstatic[0] != 0 )
		snprintf(pgmaddr,255,"%s",pgmstatic);
	      else
		snprintf(pgmaddr,255,"pgm://%s:%s",ifa->ifa_name, PGMPORT);
	    }
	}
    }
    if (ifAddrStruct!=NULL) freeifaddrs(ifAddrStruct);
    
    if(!pgmaddr[0]) {
      printf("Error: ethernet interface %s is not configured\n",ethiface);
      return(0);
    }
    return(1);
}



#define MAX 256 // max XML string size
// returns a pointer to the current position in haystack
// for subsequent calls
char *extract_xml(char *dest, char *haystack, const char *needle) {
  char *p, *pp;

  p = strstr(haystack, needle);
  if (!p) {
    fprintf(stderr, "parser cannot find \"%s\" in:\n%s\n",needle, haystack);
    return NULL;
  }
  // as you can see, i had a lot of fun with pointers when i was a kid
  pp = p; do pp+=1; while (*pp != '>'); pp++;
  p = pp; do p+=1; while (*p != '<'); *p = 0;

  snprintf(dest,MAX-1,"%s",pp);

  return(p+1);

}
int get_state(const char *match) {
  char state[MAX];
  char status[MAX];
  char speed[MAX];
  char *p;

  connect_upnp(peers[0]);
  render_upnp(peers[0],"GetTransportInfo","");
  send_upnp(peers[0]);
  recv_upnp(peers[0], 1000);
  close(peers[0]->sockfd);
  peers[0]->sockfd = 0;

  p = extract_xml(state, peers[0]->res, "CurrentTransportState"); if(!p) return(0);
  p = extract_xml(status, p, "CurrentTransportStatus");
  p = extract_xml(speed, p, "CurrentSpeed");

  //  fprintf(stderr,"#\tstate\tstatus\tspeed\n");
  //  fprintf(stderr,"TInfo:\t%s\t%s\t%s\n", state, status, speed);

  if( strcmp(match, state) == 0) return(1);
  
  return(0);
}
